from typing import ClassVar

import pydantic
import pytest

from pydantic_compat import PydanticCompatMixin


class Model(PydanticCompatMixin, pydantic.BaseModel):
    x: int = 1


def test_v1_api():
    m = Model()
    assert m.x == 1
    assert m.dict() == {"x": 1}
    assert m.json().replace(" ", "") == '{"x":1}'
    assert m.copy() == m

    assert Model.parse_raw('{"x": 2}') == Model(x=2)
    assert Model.parse_obj({"x": 2}) == Model(x=2)
    assert Model.construct(x=2) == Model(x=2)
    assert Model.validate({"x": 2}) == Model(x=2)
    assert Model.schema() == {
        "title": "Model",
        "type": "object",
        "properties": {
            "x": {"title": "X", "type": "integer", "default": 1},
        },
    }

    Model.update_forward_refs(name="name")


def test_v2_api():
    m = Model()
    assert m.x == 1
    assert m.model_dump() == {"x": 1}
    assert m.model_dump_json().replace(" ", "") == '{"x":1}'
    assert m.model_copy() == m

    assert Model.model_validate_json('{"x": 2}') == Model(x=2)
    assert Model.model_validate({"x": 2}) == Model(x=2)
    assert Model.model_construct(x=2) == Model(x=2)
    assert Model.model_validate({"x": 2}) == Model(x=2)
    assert Model.model_json_schema() == {
        "title": "Model",
        "type": "object",
        "properties": {
            "x": {"title": "X", "type": "integer", "default": 1},
        },
    }
    Model.model_rebuild(force=True)


def test_v1_attributes():
    m = Model()
    assert "x" in m.__fields__
    assert "x" in Model.__fields__
    assert "x" not in m.__fields_set__
    m.x = 2
    assert "x" in m.__fields_set__


def test_v2_attributes():
    m = Model()
    assert "x" in m.model_fields
    assert "x" in Model.model_fields
    assert "x" not in m.model_fields_set
    m.x = 2
    assert "x" in m.model_fields_set


def test_mixin_order():
    with pytest.warns(
        match="PydanticCompatMixin should appear before pydantic.BaseModel"
    ):

        class Model1(pydantic.BaseModel, PydanticCompatMixin):
            x: int = 1

    class Model2(PydanticCompatMixin, pydantic.BaseModel):
        x: int = 1


V2Config = {"populate_by_name": True, "extra": "forbid", "frozen": True}


class V1Config:
    allow_population_by_field_name = True
    extra = "forbid"
    frozen = True
    json_encoders: ClassVar[dict] = {}


@pytest.mark.parametrize("config", [V1Config, V2Config])
def test_config(config):
    class Model1(PydanticCompatMixin, pydantic.BaseModel):
        name: str = pydantic.Field(alias="full_name")

    # to make sure that populate_by_name is working
    with pytest.raises((ValueError, TypeError)):  # (v1, v2)
        m = Model1(name="John")

    class Model(PydanticCompatMixin, pydantic.BaseModel):
        name: str = pydantic.Field(alias="full_name")
        if isinstance(config, dict):
            model_config = config
        else:
            Config = config

    m = Model(name="John")

    # test frozen
    with pytest.raises((ValueError, TypeError)):  # (v1, v2)
        m.name = "Sue"

    # test extra
    with pytest.raises((ValueError, TypeError)):  # (v1, v2)
        Model(extra=1)



from unittest.mock import Mock

import pydantic
import pytest

from pydantic_compat import (
    PYDANTIC2,
    PydanticCompatMixin,
    field_validator,
    model_validator,
    root_validator,
    validator,
)


def test_v1_validator():
    mock_before = Mock()
    mock_after = Mock()

    class Model(PydanticCompatMixin, pydantic.BaseModel):
        x: int = 1

        @validator("x", pre=True)
        def _validate_x_before(cls, v):
            mock_before(v)
            return v

        @validator("x")
        def _validate_x_after(cls, v):
            mock_after(v)
            return v

    m = Model(x="2")
    mock_before.assert_called_once_with("2")
    mock_after.assert_called_once_with(2)
    assert m.x == 2


def test_v2_field_validator():
    mock_before = Mock()
    mock_after = Mock()

    class Model(PydanticCompatMixin, pydantic.BaseModel):
        x: int = 1

        @field_validator("x", mode="before")
        def _validate_x_before(cls, v):
            mock_before(v)
            return v

        @field_validator("x", mode="after")
        def _validate_x_after(cls, v):
            mock_after(v)
            return v

    m = Model(x="2")
    mock_before.assert_called_once_with("2")
    mock_after.assert_called_once_with(2)
    assert m.x == 2


def test_v1_root_validator():
    mock_before = Mock()
    mock_after = Mock()

    class Model(PydanticCompatMixin, pydantic.BaseModel):
        x: int = 1

        @root_validator(pre=True)
        def _validate_x_before(cls, v):
            mock_before(v)
            return v

        @root_validator(pre=False)
        def _validate_x_after(cls, v):
            mock_after(v)
            return v

    m = Model(x="2")
    mock_before.assert_called_once_with({"x": "2"})
    mock_after.assert_called_once_with({"x": 2})
    assert m.x == 2


@pytest.mark.xfail(PYDANTIC2, reason="not supported in pydantic v2", strict=True)
def test_v1_root_validator_with_construct():
    """Test the construct_object parameter of root_validator.

    This converts the input dict to the model object before calling the validator.
    To match the v2 behavior.  It's not supported when running on v2.  For that, just
    use model_validator(mode='after').
    """
    mock_after2 = Mock()

    class Model(PydanticCompatMixin, pydantic.BaseModel):
        x: int = 1

        @root_validator(pre=False, construct_object=True)
        def _validate_x_after2(cls, values):
            assert isinstance(values, Model)
            mock_after2(values.x)
            return values

    m = Model(x="2")
    mock_after2.assert_called_once_with(2)
    assert m.x == 2


def test_v2_model_validator():
    mock_before = Mock()
    mock_after = Mock()
    mock_after_cm = Mock()

    class Model(PydanticCompatMixin, pydantic.BaseModel):
        x: int = 1

        @model_validator(mode="before")
        def _validate_x_before(cls, v):
            mock_before(v)
            return v

        @model_validator(mode="after")
        def _validate_x_after(cls, v):
            mock_after(v)
            return v

        # this also needs to work
        @model_validator(mode="after")
        @classmethod
        def _validate_x_after_cm(cls, v):
            mock_after_cm(v)
            return v

    m = Model(x="2")
    mock_before.assert_called_once_with({"x": "2"})
    mock_after.assert_called_once_with(m)
    mock_after_cm.assert_called_once_with(m)
    assert m.x == 2



from typing import ClassVar, List, Tuple

import pytest
from typing_extensions import Literal

from pydantic_compat import BaseModel, Field


def test_field_const() -> None:
    with pytest.raises(TypeError, match="use `Literal\\['bar'\\]` instead"):
        Field(..., const="bar")  # type: ignore

    class Foo(BaseModel):
        bar: Literal["bar"] = "bar"

    with pytest.raises(ValueError, match="validation error"):
        Foo(bar="baz")  # type: ignore


@pytest.mark.parametrize("post", ["items", "length"])
@pytest.mark.parametrize("pre", ["min", "max"])
def test_field_min_max_items(pre: str, post: str) -> None:
    class Foo(BaseModel):
        bar: List[int] = Field(..., **{f"{pre}_{post}": 2})  # type: ignore

    bad_val = [1, 2, 3] if pre == "max" else [1]
    with pytest.raises((TypeError, ValueError)):  # (v1, v2)
        Foo(bar=bad_val)


def test_field_allow_mutation() -> None:
    # used in v1
    class Foo(BaseModel):
        bar: int = Field(default=1, allow_mutation=False)

        class Config:
            validate_assignment = True

    foo = Foo()
    with pytest.raises((TypeError, ValueError)):  # (v1, v2)
        foo.bar = 2


def test_field_frozen() -> None:
    # used in v2
    class Foo(BaseModel):
        bar: int = Field(default=1, frozen=True)
        model_config: ClassVar[dict] = {"validate_assignment": True}  # type: ignore

    foo = Foo()
    with pytest.raises((TypeError, ValueError)):  # (v1, v2)
        foo.bar = 2


@pytest.mark.parametrize("key", ["regex", "pattern"])
def test_regex_pattern(key: str) -> None:
    class Foo(BaseModel):
        bar: str = Field(..., **{key: "^bar$"})  # type: ignore

    Foo(bar="bar")
    with pytest.raises(ValueError):
        Foo(bar="baz")


@pytest.mark.parametrize(
    "keys",
    [
        ("min_items", "min_length"),
        ("max_items", "max_length"),
        ("allow_mutation", "frozen"),
        ("regex", "pattern"),
    ],
)
def test_double_usage_raises(keys: Tuple[str, str]) -> None:
    with pytest.raises(ValueError, match="Cannot specify both"):
        Field(..., **dict.fromkeys(keys))  # type: ignore


# not attempting unique_items yet...
# see https://github.com/pydantic/pydantic-core/issues/296
# @pytest.mark.skipif(
#     pydantic.version.VERSION.startswith("1.8"),
#     reason="pydantic 1.8 does not support unique_items",
# )
# def test_unique_items() -> None:
#     class Foo(BaseModel):
#         bar: List[int] = Field(..., unique_items=True)
#     with pytest.raises(ValueError):
#         Foo(bar=[1, 2, 3, 1])
